@@ -123,7 +123,8 @@ fn is_non_callable_value(value: &Expr) -> bool {
             | Expr::SetComp(_)
             | Expr::DictComp(_)
             | Expr::Generator(_)
-            | Expr::FString(_))
+            | Expr::FString(_)
+            | Expr::TString(_))
 }
 
 /// Generate an [`Expr`] to replace `defaultdict(default_factory=callable)` with
